import pygame, random
from pygame.locals import *
import data
import textures
from OpenGL.GL import *


class Building(object):
    """ Model for buildable buildings (markets, inns, paths, etc). """
    def __init__(self, tile, player, image=None, construction=0):
        self._player = None

        self.jeopardy = False
        self.jeopardy_count = 0

        self.tile = tile
        self.region = self.tile.region

        self.x = tile.x
        self.y = tile.y
        self.construction = construction

        self.storage = {"fish":0, "ice":0, "wood":0, "crystal":0, 'snow':0}

        self.select_surface = False

        self.tiles_in_range = {}

        self.work_force = 1
        self.workers = set()

        self.units = set()

        self.units_in_building = set()

        self.player = player

        if hasattr(self, "smoke_start"):
            self.smoke = [(self.smoke_start,40)]
            self.smoke_counter = data.get_ticks()

        self.player_color_surface = None

        self.creating_worker = 0
        self.start_time = 0

        #if not image:
            #image = self.image

        #self.image = textures.Texture("data/"+image+".png")
        if image:
            self.image_name = image

    def _get_image(self):
        return textures.Texture.get("data/"+self.image_name+".png")
    image = property(_get_image)


    def draw(self):
        x = self.x*32
        y = self.y*32

        x -= data.view.x
        y -= data.view.y

        data.draw_buffer.append(data.Rendered(self.image, x,
                (data.view.h-self.size*32)-y, z=3))
        if self.player:
            if self.jeopardy:
                data.draw_buffer.append(data.Rendered(data.jeopardy_marker, x+self.size*32-32,
                    (data.view.h-self.size*32)-y, z=4))
            else:
                data.draw_buffer.append(data.Rendered(data.marker, x+self.size*32-32,
                    (data.view.h-self.size*32)-y, colorize=self.player.glcolor, z=4))


    def _set_player(self, player):
        if player:
            self._player = player
            for u in self.units:
                u.player = player
            #size = self.size*32
            #self.player_color_surface = pygame.Surface((size+20, size-20))
            #self.player_color_surface.fill((255,0,255))
            #self.player_color_surface.set_colorkey((255,0,255))
            #pygame.draw.ellipse(self.player_color_surface, self.player.color, self.player_color_surface.get_rect())
            #self.player_color_surface.set_alpha(100)
    def _get_player(self):
        return self._player
    player = property(_get_player, _set_player)


    def draw_fluf(self):
        # Player color.
        pass
        #if self.player != self.game.player:
            ## It's an opponent. Draw the color.
            #gui_x, gui_y =  (self.x*self.game.tile_size, self.y*self.game.tile_size)
            #blit_x = gui_x - self.game.display.view_x
            #blit_y = gui_y - self.game.display.view_y
            #self.game.screen.blit(self.player_color_surface, (blit_x-10,blit_y+20))

        #else:
            #if self.creating_worker > 0:
                #pos = (self.x*self.game.tile_size-self.game.display.view_x+self.game.tile_size*self.size/2-self.creating_worker/2,
                    #self.y*self.game.tile_size-self.game.display.view_y+self.game.tile_size*self.size)
                #pygame.draw.line(self.game.screen, (0,0,200), pos,
                        #(pos[0]+self.creating_worker,pos[1]), 2)

    def draw_smoke(self):
        # Smoke.
        if hasattr(self, "smoke_start"):
            #print self.smoke
            #print len(self.smoke)
            old_smoke = self.smoke
            self.smoke = set()
            for (x,y), life_left in old_smoke:
                dx = self.x*32+x - data.view.x
                dy = self.y*32+y - data.view.y

                if life_left:
                    opc = life_left/60.0
                    s = life_left/60.0
                else:
                    opc = 0
                    s = 0

                s = 1-s

                s += 0.7

                glPushMatrix()
                glColor4f(1, 1, 1, opc)
                glTranslatef(dx, data.view.h-dy, 0)
                glScalef(s,s,1)
                data.smoke.render()
                glPopMatrix()


                #if data.get_ticks() > self.smoke_counter+30:
                a = random.randint(0,1)
                a -= 0.5
                a = max(0,a)
                x += a
                y -= 0.5
                life_left -= 1
                if life_left > 0:
                    self.smoke.add(((x,y), life_left))

            if self.units_in_building or self.creating_worker:
                if data.get_ticks() > self.smoke_counter+100:
                    self.smoke_counter = data.get_ticks()
                    sx, sy = self.smoke_start
                    sx += random.randint(-2,2)
                    self.smoke.add(((sx, sy),60))


    def can_enter(self, u):
        return False


    def get_image(self):
        """ Gets the correct image to display. """
        if self.construction < 100:
            return self.construction_image
        else:
            return self.image


    def draw_selection(self):
        """ The grey circle designating it's range and stats. """
        pass
        #gui_x, gui_y =  (self.x * self.game.tile_size,self.y * self.game.tile_size)

        #blit_x = gui_x - self.game.display.view_x
        #blit_y = gui_y - self.game.display.view_y


        #if self.game.ctrl.info_mode:
            ##pygame.draw.circle(self.game.screen, (195,195,195), (blit_x, blit_y),
                    ##int(self.range*self.game.tile_size), 2)


            #pygame.draw.rect(self.game.screen, (250,250,250),
                    #((blit_x-self.range*self.game.tile_size,
                    #blit_y-self.range*self.game.tile_size),
                    #(self.game.tile_size*self.range*2,
                    #self.game.tile_size*self.range*2)), 6)

            #text = ""
            #for r in self.storage.items():
                #if r[1] != 0:
                    #if self.construction < 100:
                        #text = text + r[0]+": "+str(r[1])+"/"+str(self.cost[r[0]])+"   "
                    #else:
                        #text = text + r[0]+": "+str(r[1])+"   "
            #text = text + "  Workforce: %d" % self.work_force
            #text = self.game.small_font.render(text, True, (50,50,50))

            #pos = (self.x*self.game.tile_size-self.game.display.view_x+self.game.tile_size*self.size/2-text.get_width()/2-1,
                    #self.y*self.game.tile_size-self.game.display.view_y+self.game.tile_size*self.size)

            #self.game.screen.blit(text,pos)


        #text = self.game.small_font.render("+  INCREASE WORK FORCE       -  DECREASE WORK FORCE", True, (100,100,100))
        #pos = (5, self.game.view.h-35)
        #self.game.screen.blit(text,pos)


    def draw_stats(self):
        #TODO: Optimize this! make it only update stats when somthing changes!
        pass
        #if self.jeopardy:
            #count = self.jeopardy_count+30000-self.game.get_ticks()
            #if count < 1000: count = 1000
            #text = "Takeover countdown: " +\
                    #str(count/1000)
            #color = (234,193,0)
            #text = self.game.small_font.render(text, True, color)
            #pos = (self.x*self.game.tile_size-self.game.display.view_x+self.game.tile_size*self.size/2-text.get_width()/2,
                    #self.y*self.game.tile_size-self.game.display.view_y+self.game.tile_size*self.size)
            #self.game.screen.blit(text,pos)

        #elif self.game.ctrl.info_mode and self.game.display.holding_building != self and self.player == self.game.player:
            #text = ""
            #color = (150,150,150)
            #if self.construction < 100:
                #for r in self.storage.items():
                    #if r[1] != 0:
                        #text = text + r[0]+": "+str(r[1])+"/"+str(self.cost[r[0]])+"   "
            #else:
                #for r in self.storage.items():
                    #if r[1] != 0:
                        #text = text + r[0]+": "+str(r[1])+"   "

            #demand_size = 0
            #for d in self.demand.items():
                #amount = d[1] - self.storage[d[0]]
                #if amount > 0:
                    #demand_size += amount
            #if demand_size == 0 and not self.construction < 100:
                #text = "FULL!"
                #color = (234,193,0)
            ##else:
                ##color = (150,150,150)
                ##text = text + str(self.work_force)
            #text = self.game.small_font.render(text, True, color)

            #pos = (self.x*self.game.tile_size-self.game.display.view_x+self.game.tile_size*self.size/2-text.get_width()/2,
                    #self.y*self.game.tile_size-self.game.display.view_y+self.game.tile_size*self.size)

            #self.game.screen.blit(text,pos)


    def is_in_range(self, obj):
        if self.construction < 100:
            return False

        in_range = True


        if obj.x <= self.x-self.range:
            in_range = False
        if obj.x >= self.x+self.range:
            in_range = False
        if obj.y <= self.y-self.range:
            in_range = False
        if obj.y >= self.y+self.range:
            in_range = False

        return in_range


    def check_const(self):
        if self.construction < 100:
            for r in self.cost.items():
                if self.storage[r[0]] < r[1]:
                    return
            self.construction = 100

            for r in self.cost.items():
                self.storage[r[0]] -= r[1]