from random import *
import pygame, os
from pygame.locals import *
import ctrl, data
from ConfigParser import ConfigParser, NoOptionError

try:
    from networking import MasterIgloo as Igloo
    from networking import MasterUnit as Unit
except ImportError:
    from buildings.igloo import Igloo
    from unit import Unit


from buildings.landobstruction import LandObstruction
from OpenGL.GL import *
import zipfile

from cStringIO import StringIO
from textures import Texture


class Map:
    def __init__(self):
        self.name = "Unnamed"
        self.tiles = {}
        self.size = (0,0)
        #self.size = 150
        self.borders = set()
        self.unit_nodes = NodeSet(7)
        self._dynamic_objects = []
        self.loaded_plugins = set()

    @classmethod
    def load(cls, name):
        map = cls()
        z = zipfile.ZipFile(os.path.join("maps", name+".zip"), "r")
        settingsfile = z.read("settings.ini")
        config = map.config = ConfigParser()
        config.readfp(StringIO(settingsfile))


        # Regions, buildings, and resources.
        simage = z.read(config.get("map", "resources"))
        simage = pygame.image.load(StringIO(simage))
        rimage = z.read(config.get("map", "regions"))
        rimage = pygame.image.load(StringIO(rimage))
        map.size = (rimage.get_width(), rimage.get_height())

        # Regions
        map.regions = {}
        for x in range(map.size[0]):
            for y in range(map.size[1]):
                rn = rimage.get_at((x,y))[:3]
                if not map.regions.has_key(rn):
                    map.regions[rn] = Region(rn)
                region = map.regions[rn]
                t = Tile("snow", x, y, region)
                map.tiles[(x,y)] = t

        # Buildings
        for n,pos in config.items("igloos"):
            x,y = pos.split(",")
            map.add_igloo(int(x),int(y))

        # Resources
        for x in range(map.size[0]):
            for y in range(map.size[1]):
                t = map.tiles[(x,y)]
                if map.pos_is_open_check((x,y)):
                    c = simage.get_at((x,y))
                    r = None
                    if c == (0,255,0,255):
                        image = data.trees[randint(0,3)]
                        Resource(t, "tree", 20, image)
                    elif c == (0,255,255,255):
                        Resource(t, "fish", 20, data.fish)

        data.map = map

        # For networking the units and buildings need to know their IDs.
        for id, u in enumerate(data.units):
            u.unit_id = id
        for id, b in enumerate(data.buildings):
            b.building_id = id

        # Plugins
        # TODO: Make this only run if there are plugins!
        if data.THIS_IS_SERVER:
            plugins = config.get("map", "required_plugins").split(",")
            for p in plugins:
                p,v = p.split("-v")
                try:
                    pl = __import__("plugins."+p, None, None, "plugins")
                except ImportError:
                    raise ImportError("Could not find plugin %s"%p)
                if int(v) > pl.__version__:
                    raise ImportError("%s plugin is too old. Required version is %s"%(p,v))
                map.loaded_plugins.add(pl.init(config))

        #map.plugins = set()
        #for _,params in config.items("plugins"):
            #params = params.split(",")
            #p = params.pop(0)
            #p,clsname = p.split(".")
            #pl = __import__("plugins."+p, None, None, "plugins")
            #cls = pl.__dict__[clsname]
            #map.plugins.add(cls(*params))

        return map

    def add_igloo(self,x,y):
        if not data.THIS_IS_SERVER:
            from networking import SlaveUnit as U
            from networking import SlaveIgloo as I
        else:
            U = Unit
            I = Igloo

        new_building = I(self.tiles[(x,y)], player=None, construction=100)

        root_tile = self.tiles[(x,y)]
        if root_tile.region.igloo:
            raise ValueError("You can only have 1 igloo per region!")
        tiles = []
        for xx in range(new_building.size):
            for yy in range(new_building.size):
                c = self.tiles[x+xx,y+yy]
                if c.building == None and c.type == "snow" and not c.resource:
                    tiles.append(c)
                else:
                    raise ValueError("building already there!")

        data.buildings.append(new_building)
        root_tile.building_root = True
        for c in tiles:
            c.building = new_building

        # Add units.
        for i in xrange(self.config.getint("igloo_settings", "workers")):
            u = U((new_building.x+1,new_building.y+1),
                    new_building, "worker")
            u.warmth = u.max_warmth = self.config.getint("workers", "max_warmth")
            data.units.append(u)

        for i in xrange(self.config.getint("igloo_settings", "snowballers")):
            u = U((new_building.x+1,new_building.y+1),
                    new_building, "snowballer")
            u.warmth = u.max_warmth = self.config.getint("snowballers", "max_warmth")
            data.units.append(u)


    def pos_is_open_check(self, pos):
        if not self.tiles.has_key(pos): return False
        t = self.tiles[pos]
        if not t.building and t.type == "snow" and not t.resource:
            return True
        return False

    def place_landobstruction(self, pos, name):
        new_building = LandObstruction(pos, name)

        root_tile = self.tiles[pos]

        tiles = set()
        for x in range(new_building.size):
            for y in range(new_building.size):
                c = self.tiles[pos[0]+x,pos[1]+y]
                if c.building == None and c.type == "snow" and not c.resource:
                    tiles.add(c)
                else:
                    raise ValueError("building already there!")

        data.buildings.append(new_building)
        root_tile.building_root = True
        for c in tiles:
            c.building = new_building

    #def set_border(self, tile, tile2):
        #tile.border = True
        #tile2.border = True
        #self.borders.update(set((tile, tile2)))
        #tile.region.adjacent_regions.add(tile2.region)
        #tile2.region.adjacent_regions.add(tile.region)


    #def detect_same_borders(self, tile, tile2):
        #if tile.border != True or tile2.border != True:
            #return
        #side = None
        #if tile.x == tile2.x-1 and tile.y == tile2.y:
            #side = "left"
        #elif tile.x == tile2.x+1 and tile.y == tile2.y:
            #side = "right"
        #elif tile.x == tile2.x and tile.y == tile2.y-1:
            #side = "top"
        #elif tile.x == tile2.x and tile.y == tile2.y+1:
            #side = "bottom"

        #if side != None:
            #tile.adjacent_borders[side] = True

    #def detect_shore_line(self, tile, tile2):
        #if tile.border != True or tile2.border != True:
            #return
        #side = None
        #if tile.x == tile2.x-1 and tile.y == tile2.y:
            #side = "left"
        #elif tile.x == tile2.x+1 and tile.y == tile2.y:
            #side = "right"
        #elif tile.x == tile2.x and tile.y == tile2.y-1:
            #side = "top"
        #elif tile.x == tile2.x and tile.y == tile2.y+1:
            #side = "bottom"

        #if side != None:
            #tile.shore_line[side] = True

    #def calc_regions(self):
        #self.detect_edges(self.tiles, "region", self.set_border, "border")
        ##self.detect_edges(self.tiles, "region", self.detect_same_borders,
                ##compare_value=True)
        ##self.detect_edges(self.tiles, "type", self.detect_shore_line, check_sides=4)

        ##for tile in self.tiles.values():
            ##if tile.type == "water":
                ##tile.water_shore()

    #def detect_edges(self, objects, attr, action, exception=None,
            #pass_excp_on=True, compare_value=False, check_sides=8):
        #""" Loops through objects and compares differences between attr (for
            #tiles, objects must be a dict with (x,y) tuples as keys). When a
            #difference is found, calls action and passes the mismatch objects.
            #Finds the attribute that the exception is set to (if set) and passes
            #if the exception equals pass_excp_on.
            #Some objects will probably be passed to the action multiple times.
            #WARNING: This function is slow! Don't do run it too often!"""
        #for obj in objects.items():
            #if exception and getattr(obj[1], exception) == pass_excp_on:
                #continue

            #x,y = obj[0]
            #thisattr = getattr(obj[1], attr)
            #if check_sides == 8:
                #to_check = ((x-1,y), (x+1,y), (x,y+1), (x,y-1), (x+1,y+1),
                        #(x-1,y-1), (x-1,y+1), (x+1,y-1))
            #else:
                #to_check = ((x-1,y), (x+1,y), (x,y+1), (x,y-1))
            #for item in to_check:
                #v = objects.get(item)
                #if not v: continue
                #thatattr = getattr(v, attr)
                #if thisattr != None and thatattr != None:
                    #if compare_value == False:
                        #if thatattr != thisattr:
                            #action(obj[1], v)
                    #elif compare_value == True:
                        #if thatattr == thisattr:
                            #action(obj[1], v)


    def add_dynamic_object(self, object):
        # TODO alert clients
        object.dmo_id = len(self._dynamic_objects)
        self._dynamic_objects.append(object)

    def remove_dynamic_object(self, object):
        # We can't just remove it because it would mess up looking up other
        # objects by their id.
        object.hidden = True
        #self._dynamic_objects.remove(object)


    def draw(self):
        for r in data.resources:
            x = r.x*32
            y = r.y*32
            if x+64 > data.view.x and x < data.view.x+data.view.w\
                    and y+64 > data.view.y and y < data.view.y+data.view.h+64:

                x -= data.view.x
                y -= data.view.y

                data.draw_buffer.append(data.Rendered(r.image, x,
                        (data.view.h-32)-y, z=1))

        for o in self._dynamic_objects:
            if o.hidden: continue
            x,y = o.position
            x = x*32
            y = y*32
            t = Texture.get(o.image_name)
            w,h = t.width, t.height
            if x+w > data.view.x and x < data.view.x+data.view.w\
                    and y+h > data.view.y and y < data.view.y+data.view.h+h:
                x -= data.view.x
                y -= data.view.y
                data.draw_buffer.append(data.Rendered(t, x, (data.view.h-32)-y, z=1))


class NodeSet:
    def __init__(self, num_tiles):
        self.node_size = num_tiles
        self.nodes = {}

    def find_current_node(self, tilex,tiley):
        i = (tilex//self.node_size, tiley//self.node_size)
        if not self.nodes.has_key(i):
            node = set()
            self.nodes[i] = node
            return node
        else:
            return self.nodes[i]

    def get_nodes_around_tile(self, pos):
        x = int(pos[0])//self.node_size
        y = int(pos[1])//self.node_size
        idxs = ((x, y), (x-1, y), (x+1, y),
                (x, y+1), (x-1, y+1), (x+1, y+1),
                (x, y-1), (x-1, y-1), (x+1, y-1))
        for idx in idxs:
            if self.nodes.has_key(idx):
                yield self.nodes[idx]

    def get_objs_around_tile(self, pos):
        for node in self.get_nodes_around_tile(pos):
            for obj in node:
                yield obj


class Tile:
    """ A map tile. """
    def __init__(self, type, x, y, region):
        self.x = x
        self.y = y
        self.type = type
        self.building = None
        self.building_root = False
        self.resource = None
        self.region = region
        self.region.tiles.add(self)
        #self.border = False
        #self.adjacent_borders = {"top":False, "bottom":False,
                #"left":False, "right":False}
        #self.shore_line = {"top":False, "bottom":False,
                #"left":False, "right":False}


    def is_walkable(self):
        return self.type == "snow"

class Region(object):
    def __init__(self, name):
        self.name = name
        self._player = None
        self.tiles = set()
        self.resources = set()
        self.adjacent_regions = set()
        self._igloo = None


    def _get_igloo(self):
        if not self._igloo:
            for b in data.buildings:
                if b.name == "igloo" and b.region == self:
                    self._igloo = b
                    break
        return self._igloo
    igloo = property(_get_igloo)

    def _set_player(self, value):
        #if value:
            #print value.name
        old_player = self._player
        self._player = value

        for t in self.tiles:
            if t.resource:
                if old_player: old_player.open_resources.remove(t.resource)
                self._player.open_resources.add(t.resource)
            if t.building_root and t.building and t.building.name == "igloo":
                t.building.player = self._player

        for player in data.players.values():
            if player and player.type == "ai":
                player.ai.setup()

    def _get_player(self):
        return self._player
    player = property(_get_player, _set_player)


class Resource:
    def __init__(self, tile, type, quantity, image=None):
        self.tile = tile
        self.x, self.y = tile.x, tile.y
        data.resources.add(self)
        tile.resource = self
        tile.region.resources.add(self)
        self.type = type
        self.image = image
        self.quantity = quantity

    def replenish(self):
        self.quantity = randint(20,50)
