import sprite
import textures
import os
from OpenGL.GL import *
import data
from random import *
import math, sys
import snowball
import events



class Unit(object):
    def __init__(self, pos, igloo, type, state="inbuilding"):
        self.direction = None

        self.x,self.y = pos

        self.current_frame = randint(0, 10)
        self.last_tick = 0

        self.selected = False
        self.type = type
        self.level = 1
        self.max_warmth = 10
        self.warmth = self.max_warmth
        self.prv_node = None

        self.gather_time = 20000
        self.create_snowball_time = 3000

        self.igloo = igloo
        self.igloo.units.add(self)

        self.holding = None
        self.current_job = None

        self._go_to_building = None

        self.state = state
        if state == "inbuilding":
            self.go_to_building = igloo
            igloo.units_in_building.add(self)

        self.targeted_unit = None


        self.move_time = 0
        self.walk_speed = 800.
        self.offset = 0
        self.moving = False

        # Direction that the unit is moving in
        self.direction = (0,-1)

        self.target_resource_tile = None

        # Used so that the penguin doesn't get stuck going back and forth.
        self.last_tiles = []

        self.counter = 0

        self.snowballs = 1

        self.gathering_from = None

        self.going_for = ""

        self.fail_time = 0

        self.run_mode = False

        # Plugins applied to this penguin.
        self.plugins = set()
        self._emblems = set()

    def add_emblem(self, emblem):
        self._emblems.add(emblem)

    def remove_emblem(self, emblem):
        self._emblems.remove(emblem)

    def _set_go_to_building(self, building):
        if self.state != "entering" and (self.state != "inbuilding" or not self._go_to_building):
            self._go_to_building = building
    go_to_building = property(lambda self: self._go_to_building, _set_go_to_building)


    def draw(self):
        if self.state == "frozen":
            t = data.penguins["frozen"]
            to = None
        elif self.state in ["unloading", "throwing", "getting_snowball",
                "gathering"]:
            t = data.penguins[self.state].images[(self.current_frame,
                        self.direction)]
            to = data.penguins[self.type+self.state].images[(self.current_frame,
                        self.direction)]
        elif self.holding and self.type == "worker":
            t = data.penguins["walkfish"].images[(self.current_frame,
                        self.direction)]
            to = data.penguins["workerwalkfish"].images[(self.current_frame,
                        self.direction)]
        else:
            if self.moving:
                t = data.penguins["walk"].images[(self.current_frame,
                        self.direction)]
                to = data.penguins[self.type+"walk"].images[(self.current_frame,
                        self.direction)]
            else:
                t = data.penguins["walk"].images[("idle", self.direction)]
                to = data.penguins[self.type+"walk"].images[("idle", self.direction)]
        x = self.x*32
        y = self.y*32

        offset = self.offset
        if self.moving:
            offset -= 32

        x = x - data.view.x + self.direction[0]*offset
        y = y - data.view.y + self.direction[1]*offset

        if self.selected:
            # Draw selection.
            data.draw_buffer.append(data.Rendered(data.penguinselection, x,
                (data.view.h-32)-y, z=1))

        # Draw penguin
        self.draw_x = x
        self.draw_y = (data.view.h-32)-y
        data.draw_buffer.append(data.Rendered(t, x,
                (data.view.h-32)-y, z=1))

        if to:
            # Draw penguin outfit.
            if self.igloo.player: color = self.igloo.player.glcolor
            else: color = (1,1,1)
            data.draw_buffer.append(data.Rendered(to, self.draw_x, self.draw_y,
                    z=1, colorize=color))


        for e in self._emblems:
            e.draw(x, (data.view.h-32)-y)


        # Warmth bar.
        if self.state != "frozen" and self.warmth < self.max_warmth:
            r = lambda w: max(min(abs((w/float(self.max_warmth)*2)-2), 1), 0)
            g = lambda w: max(min(w/float(self.max_warmth)*2, 1), 0)

            x += 32
            y += 28

            height = self.warmth*20/self.max_warmth

            glPushMatrix()
            glColor3f(r(self.warmth),g(self.warmth),0)

            glTranslatef(x, (data.view.h)-y, 0)
            glBegin(GL_QUADS)
            glVertex2f(0, 0)
            glVertex2f(2, 0)
            glVertex2f(2, height)
            glVertex2f(0, height)
            glEnd()

            glColor3f(1,1,1)
            glPopMatrix()


    def animate(self, dt):
        #TODO: make the rate of penguin walking based on time.

        #if not self.last_tick:
            #self.last_tick = dt

        #if self.last_tick+2 <= dt:
            #print dt
            self.current_frame += 1
            if self.current_frame > 9:
                self.current_frame = 0

                #if self.moving:
                    #x = self.x*32
                    #y = self.y*32

                    #offset = self.offset
                    #if self.moving:
                        #offset -= 32

                    #x += self.direction[0]*offset
                    #y += self.direction[1]*offset

                    #y += 32

                    #data.footimprents.add(x,y)

                # BAD HACK to uload penguins
                if self.state == "unloading":
                    # Finished unloading.
                    self.state = "stray"
                    if data.THIS_IS_SERVER:
                        self.current_job.workers.remove(self)
                        self.current_job.storage[self.holding] += 1
                        self.current_job = None
                    self.holding = None


    def pos_in_dir(self, direction):
        """ Returns the next position in this direction relative to the
            penguin. """
        return (self.x+direction[0], self.y+direction[1])
    def try_direction(self, direction):
        """ Check if the penguin can move to this direction. """
        pos = self.pos_in_dir(direction)
        try:
            tile = data.map.tiles[pos]
        except KeyError:
            return False
        if not tile.is_walkable():
            return False


        if self.state != "entering":
            if tile.building:
                # Penguins can walk on the top row of a building.
                if not tile.building.y == pos[1]:
                    return False

            for u in data.map.unit_nodes.get_objs_around_tile((self.x,self.y)):
                if u is not self:
                    if (u.x,u.y) == pos:
                        return False
        else:
            self.state = "inbuilding"

        if tile.resource:
            return False

        if len(self.last_tiles) >= 30:
            self.last_tiles.pop(0)

        if (tile.x,tile.y) in self.last_tiles:
            return False


        return True


    def move(self):
        # Skip the check if the penguin is moving.
        if not self.moving:
            # Check with the AI to see if we want to move somewhere.
            if data.get_ticks() > self.fail_time+randint(200, 400):
                r = self.check_direction()
                if r is False:
                    self.fail_time = data.get_ticks()
        else:
            time_passed = data.get_ticks()-self.counter
            self.offset = int(time_passed/self.walk_speed*32)
            if self.offset >= 32:
                self.offset = 0
                self.check_direction()

    def check_direction(self):
        direction = self.find_path()
        if direction:
            # Move penguin.
            self.last_tiles.append((self.x,self.y))
            self.x,self.y = self.pos_in_dir(direction)

            events.fire(events.PENGUIN_MOVE, self, self.x, self.y)

            # Update position in node.
            current_node = data.map.unit_nodes.find_current_node(self.x,self.y)
            if not (self.prv_node is current_node):
                if self.prv_node:
                    self.prv_node.remove(self)
                current_node.add(self)
                self.prv_node = current_node

            self.direction = direction
            self.moving = True
            self.counter = data.get_ticks()
            return True
        else:
            self.moving = False
            return direction


    def find_path(self):
        if self.offset != 0:
            # Return if penguin is currently moving.
            return None

        direction = self.stray()
        if not direction:
            return direction

        turn_right_map = {
                (0,1)  :(-1,1),
                (-1,1) :(-1,0),
                (-1,0) :(-1,-1),
                (-1,-1):(0,-1),
                (0,-1) :(1,-1),
                (1,-1) :(1,0),
                (1,0)  :(1,1),
                (1,1)  :(0,1)
                }
        turn_left_map = {
                (-1, -1): (-1, 0),
                (-1, 0): (-1, 1),
                (-1, 1): (0, 1),
                (0, -1): (-1, -1),
                (0, 1): (1, 1),
                (1, -1): (0, -1),
                (1, 0): (1, -1),
                (1, 1): (1, 0)
                }

        turn_left = turn_left_map[direction]
        turn_right = direction
        while True:
            if self.try_direction(turn_right):
                direction = turn_right
                break
            if self.try_direction(turn_left):
                direction = turn_left
                break
            turn_right = turn_right_map[turn_right]
            # Check to see if we have exahsted all posibilities.
            if turn_right == turn_left:
                self.last_tiles = []
                return False
            turn_left = turn_left_map[turn_left]
        return direction





    def stray(self):
        """ AI for penguins.
            This code is a MESS!!! Read at your own risk! Don't say I didn't
            warn you... """

        time = data.get_ticks()

        if self.warmth == 0:
            if data.map.tiles[(self.x,self.y)].building_root == True:
                self.warmth = 1

        if self.warmth < 1:
            self.state = "frozen"
            return False


        # can't interrupt these states.
        if self.state == "unloading":
            return None
        if self.state == "entering":
            self.last_tiles = []
            self.counter = data.get_ticks()
            self.go_to_building.units_in_building.add(self)
            self.go_to_building.unit_enter(self)
            return (1,1)
        if self.state == "inbuilding":
            return self.go_to_building.in_building(self)
        if self.state == "throwing":
            # Throw it!
            if self.current_frame < 9:
                self.state = "throwing"
                return None
            else:
                # All done throwing.
                self.state = "stray"
                self.targeted_unit = None


        # Go to igloo if too cold.
        if self.warmth < 5:
            # clear job.
            self.state = "stray"
            self.gathering_from = None
            self.target_resource_tile = None
            if self.current_job:
                if hasattr(self.current_job, "workers"):
                    self.current_job.workers.remove(self)
                self.current_job = None
            if self.igloo.can_enter(self):
                self.go_to_building = self.igloo

        else:
            # If penguin gets cold, these states can be interrupted.
            if self.state == "getting_snowball":
                if self.counter+self.create_snowball_time <= data.get_ticks():
                    self.snowballs += 1
                    events.fire(events.PENGUIN_RELOAD, self)
                    self.state = "stray"
                else:
                    return None
            if self.state == "gathering":
                self.last_tiles = []
                if self.gathering_from.resource.quantity == 0:
                    # Resource ran out while gathering.
                    self.gathering_from = None
                    self.target_resource_tile = None
                    self.state = "stray"
                    return self.stray_at_igloo()
                if self.counter+self.gather_time <= data.get_ticks():
                    # Finnished gathering.
                    self.gathering_from.resource.quantity -= 1
                    self.holding = self.gathering_from.resource.type
                    self.gathering_from = None
                    self.target_resource_tile = None
                    self.state = "stray"
                    return self.move_to_building(self.current_job)
                else:
                    return None

        if not self.igloo.player:
            return self.stray_at_igloo()

        if self.type == "snowballer" and self.igloo.jeopardy == True:
            self.go_to_building = self.igloo

        if self.go_to_building:
            if (self.x, self.y) == (self.go_to_building.x,self.go_to_building.y):
                if self.go_to_building.can_enter(self):
                    self.state = "entering"
                    return
                else:
                    self.go_to_building = None
        if self.go_to_building:
            return self.move_to_building(self.go_to_building)

        # What is this for??? For some reason it must be before next if block!
        if self.type == "snowballer" and\
        data.map.tiles[(self.x,self.y)].building_root and\
        data.map.tiles[(self.x,self.y)].building.player is self.igloo.player\
        or data.map.tiles[(self.x,self.y)].building:
            return self.stray_at_igloo(True)

        if self.type == "worker":
            if not self.go_to_building:
                if not self.current_job:
                    # Get a job.
                    self.last_tiles = []
                    self.current_job = self.igloo
                    self.igloo.workers.add(self)

                direction = self.work_at_job(self.current_job)
                if direction:
                    return direction
                if self.state == "gathering" or self.state == "unloading":
                    return None
                return self.stray_at_igloo()


        elif self.type == "snowballer":
            # We don't want it to stop if in run_mode.
            if not self.run_mode:
                if self.snowballs:
                    units_in_range = self.check_enemy_units_in_range()
                    if units_in_range:
                        if not self.targeted_unit:
                            snowballers = [u for u in units_in_range if
                                    u.type == "snowballer"]

                            if snowballers:
                                units_in_range = snowballers

                            for u in units_in_range:
                                if data.map.tiles[(self.x,self.y)].building_root:
                                    continue

                                # Make sure not taking over building belonging to this penguin.
                                if u.warmth < 5:
                                    c = False
                                    for un in u.igloo.units_in_building:
                                        if un.player == self.igloo.player:
                                            c = True
                                            break
                                    if c:
                                        continue

                                if not self.targeted_unit:
                                    self.targeted_unit = u
                                else:
                                    if ((u.x-self.x)**2 + (u.y-self.y)**2)**.5 <\
                                            (abs(self.targeted_unit.x-self.x)**2 +\
                                            abs(self.targeted_unit.y-self.y)**2)**.5:
                                        self.targeted_unit = u
                        # Putting this in an else makes it so they will move farward
                        # before throwing again.
                        else:
                            if not self.targeted_unit in units_in_range:
                                self.targeted_unit = None
                                return None
                            else:
                                self.throw_snowball(self.targeted_unit)
                                self.targeted_unit = None
                                return None
                else:
                    # Get a snowball.
                    self.state = "getting_snowball"
                    self.counter = data.get_ticks()
                    return None



            if self.current_job:
                j = self.current_job

                if self.is_in_range(j):
                    # Reached destination. It's now okay to through and gather
                    # snowballs again.
                    self.run_mode = False

                    # Stray around the current job position.
                    self.last_tiles = []
                    will_stray = randint(1,3)
                    if will_stray == 1:
                        d_x = randint(-1, 1)
                        d_y = randint(-1, 1)
                        if not d_x and not d_y:
                            return False

                        return (d_x,d_y)
                    return False
                else:
                    return self.find_next_tile_to_pos(*j)
            else:
                # Move to alert tiles.
                ts = []
                for t in self.igloo.player.alert_tiles.items():
                    x,y = t[0]
                    if self.igloo.player.snowballer_alert_range >= ((x-self.x)**2 + (y-self.y)**2)**.5 and t[1] != 0:
                        ts.append(t)

                if ts:
                    best_t = None
                    for t in ts:
                        x,y = t[0]
                        leng = ((x-self.x)**2 + (y-self.y)**2)**.5
                        if leng and (not best_t or leng < best_t[1]):
                            best_t = (t,leng)

                    #if best_t[1] > 4:
                    if best_t:
                        return self.find_next_tile_to_pos(x=best_t[0][0][0],
                                y=best_t[0][0][1])
                    else:
                        return None


        # No job availible. Stray around home igloo.
        return self.stray_at_igloo()


    def throw_snowball(self, target):
        # Throw snowball.
        s = snowball.Snowball.throw_at_target(self, target)
        data.snowballs.add(s)
        self.snowballs -= 1
        self.state = "throwing"
        self.direction = self.direction_to_tile(target.x, target.y)
        self.current_frame = 1
        return s

    def direction_to_tile(self, tx, ty):
        """
        returns the direction from this penguin to the given tile coordinates.
        """
        # FIXME  This could perform better for tiles that are not immediately
        # adjacent.  (It is WAY to biased tward the diagonal.)
        if tx > self.x:
            x = 1
        elif tx < self.x:
            x = -1
        else:
            x = 0
        if ty > self.y:
            y = 1
        elif ty < self.y:
            y = -1
        else:
            y = 0
        return (x, y)

    def stray_at_igloo(self, will_stray=False):
        if self.igloo.is_in_range(self):
            if not will_stray:
                will_stray = randint(1,5)
            if will_stray == 1 or will_stray == True:
                d_x = randint(-1, 1)
                d_y = randint(-1, 1)
                if d_x == 0 and d_y == 0:
                    return False

                return (d_x,d_y)
            return False
        else:
            return self.move_to_building(self.igloo)



    def set_job(self, pos, run_mode=False):
        if self.state == "entering" or self.state == "inbuilding":
            return
        if self.type == "snowballer":
            if data.map.tiles[pos].building:
                if not self in data.map.tiles[pos].building.units_in_building:
                    self.last_tiles = []
                    self.go_to_building = None
                    self.current_job = None
                    self.go_to_building = data.map.tiles[pos].building
            else:
                self.last_tiles = []
                self.go_to_building = None
                self.current_job = pos
            self.run_mode = run_mode

    def disban(self):
        self.selected = False
        self.go_to_building = None
        self.current_job = None


    def work_at_job(self, building):
        demand = self.going_for
        if not self.holding:
            demanded_resources = self.check_demand(building)
            already_got_job = False
            if demanded_resources.has_key(self.going_for):
                if demanded_resources[self.going_for]:
                    already_got_job = True

            if not already_got_job:
                ld = 0
                for d in demanded_resources.items():
                    if d[1] > ld:
                        ld = d[1]
                        self.going_for = demand = d[0]
        else:
            demand = self.holding

        if demand:
            if self.holding:
                # Got a resource, go dump it.
                direction = self.move_to_building(building)
                next_pos = (direction[0]+self.x, direction[1]+self.y)
                if next_pos == (building.x,building.y):
                    self.state = "unloading"
                    self.current_frame=1
                    return
                return direction
            else:
                # Go get resource.
                direction = self.move_to_resource(demand)
                if self.state == "gathering":
                    return False
                if not direction:
                    # Unable to gather resource.
                    self.current_job = None
                    building.workers.remove(self)
                    return False
                return direction
        else:
            self.current_job = None
            building.workers.remove(self)
            return False


    def check_demand(self, b):
        if b.construction < 100:
            demand = b.cost
        else:
            demand = b.demand
        return demand



    def move_to_building(self, building):
        return self.find_next_tile_to_object(building)


    def move_to_resource(self, resource_type):
        if self.target_resource_tile:
            # Make sure that the resource it is going after is still the same as the demand.
            if resource_type != self.target_resource_tile.resource.type:
                self.target_resource_tile = None

        # Randomly pick from the nearest resources.
        if not self.target_resource_tile:
            best_option = None
            for r in self.igloo.player.open_resources:
                if r.type == resource_type and r.quantity > 0:
                    dist = ((r.x-self.x)**2 + (r.y-self.y)**2)**.5
                    if not best_option:
                        best_option = (dist,r)
                    else:
                        if dist < best_option[0] + 6 and randint(0,1):
                            best_option = (dist,r)

            if best_option:
                self.target_resource_tile = best_option[1].tile
            else:
                # No more resources of this type availible.
                return False

        direction = self.find_next_tile_to_object(self.target_resource_tile)
        next_tile = data.map.tiles[(direction[0]+self.x, direction[1]+self.y)]

        if next_tile is self.target_resource_tile:
            if next_tile.resource.quantity > 0:
                self.gathering_from = next_tile
                self.counter = data.get_ticks()
                self.state = "gathering"
                self.direction = direction
                return None
            else:
                # Resources ran out here.
                self.target_resource_tile = None
                return False
        else:
            return direction


    def find_next_tile_to_object(self, object):
        x, y = object.x, object.y
        return self.find_next_tile_to_pos(x, y)

    def find_next_tile_to_pos(self, x, y):
        dx = x - self.x
        dy = y - self.y
        # Normalize...
        d = (dx**2+dy**2)**.5
        dx, dy = dx*(1/d), dy*(1/d)
        # Make it so that the to right corner is (1,1) etc.
        dx *= 2**.5
        dy *= 2**.5

        x = y = 0
        if dx == 0.0:
            x = 0
        elif dx > 0 and dx > random():
            x = 1
        elif dx < 0 and dx < -random():
            x = -1
        if dy == 0.0:
            y = 0
        elif dy > 0 and dy > random():
            y = 1
        elif dy < 0 and dy < -random():
            y = -1

        return (x,y)


    def check_enemy_units_in_range(self):
        units_in_range = set()
        for u in data.map.unit_nodes.get_objs_around_tile((self.x,self.y)):
            if u.igloo.player is not self.igloo.player:
                if  not (u.x > self.x+7 or u.x < self.x-7)\
                and not (u.y > self.y+7 or u.y < self.y-7):
                    if u.warmth > 0:
                        if u.state != "inbuilding":
                            units_in_range.add(u)
        return units_in_range


    def detect_hit(self, snowball, x, y):
        d = 0
        hit = 0
        if x == self.x and y == self.y:
            d = 3
            self.last_tiles = []
            hit = 2
            events.fire(events.PENGUIN_DIRECT_HIT, self, d)
        elif x in range(self.x-1,self.x+1) and y in range(self.y-1,self.y+1):
            d = 1
            self.last_tiles = []
            hit = 1
            events.fire(events.PENGUIN_HIT, self, d)

        if hit and self.warmth > 0 and self.state != "inbuilding":
            x,y = snowball.start_x//32, snowball.start_y//32
            self.igloo.player.alert_tiles[(x, y)] = 3
            self.warmth -= d
            self.last_tiles = []

        if self.warmth < 1:
            self.warmth = 0
            events.fire(events.PENGUIN_FROZEN, self, snowball.penguin)

        return hit

    def is_in_range(self, (x,y)):
        in_range = True
        jrange = 4

        if x <= self.x-jrange:
            in_range = False
        if x >= self.x+jrange:
            in_range = False
        if y <= self.y-jrange:
            in_range = False
        if y >= self.y+jrange:
            in_range = False

        return in_range